@@ -10,12 +10,13 @@ from typing import Optional
 import sys
 import argparse
 import os
+import platform
 import shlex
 import shutil
 import subprocess
 import multiprocessing
 from .. import architectures
-from ..util import SilentError, uname
+from ..util import SilentError
 
 
 def make_parser():
@@ -27,7 +28,7 @@ def make_parser():
         "--arch",
         action="store",
         metavar="ARCHITECTURE",
-        default=uname.machine,
+        default=platform.machine(),
         help="Target architecture",
     )
 
@@ -103,7 +104,7 @@ def arg_fail(message):
 
 
 _GENERIC_CONFIG = [
-    "# Generic",
+    "##: Generic",
     "CONFIG_UEVENT_HELPER=n",  # Obsolete and slow
     "CONFIG_VIRTIO=y",
     "CONFIG_VIRTIO_PCI=y",
@@ -131,25 +132,25 @@ _GENERIC_CONFIG = [
     "CONFIG_EARLY_PRINTK=y",
     "CONFIG_INOTIFY_USER=y",
     "",
-    "# virtio-scsi support",
+    "##: virtio-scsi support",
     "CONFIG_BLOCK=y",
     "CONFIG_SCSI_LOWLEVEL=y",
     "CONFIG_SCSI=y",
     "CONFIG_SCSI_VIRTIO=y",
     "CONFIG_BLK_DEV_SD=y",
     "",
-    "# virt-serial support",
+    "##: virt-serial support",
     "CONFIG_VIRTIO_CONSOLE=y",
     "",
-    "# watchdog (useful for test scripts)",
+    "##: watchdog (useful for test scripts)",
     "CONFIG_WATCHDOG=y",
     "CONFIG_WATCHDOG_CORE=y",
     "CONFIG_I6300ESB_WDT=y",
-    "# Make sure debuginfo are available",
+    "##: Make sure debuginfo are available",
     "CONFIG_DEBUG_INFO_DWARF_TOOLCHAIN_DEFAULT=y",
-    "# Enable overlayfs",
+    "##: Enable overlayfs",
     "CONFIG_OVERLAY_FS=y",
-    "# virtio-fs support",
+    "##: virtio-fs support",
     "CONFIG_DAX=y",
     "CONFIG_DAX_DRIVER=y",
     "CONFIG_FS_DAX=y",
@@ -158,12 +159,15 @@ _GENERIC_CONFIG = [
     "CONFIG_ZONE_DEVICE=y",
     "CONFIG_FUSE_FS=y",
     "CONFIG_VIRTIO_FS=y",
+    "##: vsock support",
+    "CONFIG_VSOCKETS=y",
+    "CONFIG_VIRTIO_VSOCKETS=y",
 ]
 
 _GENERIC_CONFIG_OPTIONAL = [
-    "# initramfs support",
+    "##: initramfs support",
     "CONFIG_BLK_DEV_INITRD=y",
-    "# BPF stuff & useful debugging features",
+    "##: BPF stuff & useful debugging features",
     "CONFIG_BPF=y",
     "CONFIG_BPF_SYSCALL=y",
     "CONFIG_BPF_JIT=y",
@@ -180,16 +184,16 @@ _GENERIC_CONFIG_OPTIONAL = [
     "CONFIG_UPROBES=y",
     "CONFIG_UPROBE_EVENTS=y",
     "CONFIG_DEBUG_FS=y",
-    "# Required to generate memory dumps for drgn",
+    "##: Required to generate memory dumps for drgn",
     "CONFIG_FW_CFG_SYSFS=y",
     "CONFIG_FW_CFG_SYSFS_CMDLINE=y",
-    "# Graphics support",
+    "##: Graphics support",
     "CONFIG_DRM=y",
     "CONFIG_DRM_VIRTIO_GPU=y",
     "CONFIG_DRM_VIRTIO_GPU_KMS=y",
     "CONFIG_DRM_BOCHS=y",
     "CONFIG_VIRTIO_IOMMU=y",
-    "# Sound support",
+    "##: Sound support",
     "CONFIG_SOUND=y",
     "CONFIG_SND=y",
     "CONFIG_SND_SEQUENCER=y",
@@ -277,6 +281,7 @@ _GENERIC_CONFIG_OPTIONAL = [
     "# CONFIG_MSDOS_FS is not set",
     "# CONFIG_AUTOFS4_FS is not set",
     "# CONFIG_AUTOFS_FS is not set",
+    "# CONFIG_NVRAM is not set",
 ]
 
 
@@ -284,6 +289,7 @@ def do_it():
     args = _ARGPARSER.parse_args()
 
     arch = architectures.get(args.arch)
+    is_native = args.arch == platform.machine()
 
     custom_conf = []
     if args.custom:
@@ -296,7 +302,7 @@ def do_it():
 
     mod_conf = []
     if args.configitem:
-        mod_conf += ["# final config-item mods"]
+        mod_conf += ["##: final config-item mods"]
         for conf_item in args.configitem:
             if not conf_item.startswith("CONFIG_"):
                 conf_item = "CONFIG_" + conf_item
@@ -307,7 +313,7 @@ def do_it():
 
     conf = (
         _GENERIC_CONFIG_OPTIONAL
-        + ["# Arch-specific options"]
+        + ["##: Arch-specific options"]
         + arch.config_base()
         + custom_conf
         + mod_conf
@@ -324,7 +330,7 @@ def do_it():
     if args.cross_compile != "":
         cross_compile_prefix = args.cross_compile
 
-    if shutil.which(f"{cross_compile_prefix}-gcc") and arch.gccname != uname.machine:
+    if not is_native and shutil.which(f"{cross_compile_prefix}-gcc"):
         gccname = shlex.quote(f"{cross_compile_prefix}-gcc")
         archargs.append(f"CROSS_COMPILE={gccname}")
 
@@ -355,11 +361,13 @@ def do_it():
 
     # Determine if an initial config is present
     config = ".config"
+    makef = "Makefile"
 
     # Check if KBUILD_OUTPUT is defined and if it's a directory
     config_dir = os.environ.get("KBUILD_OUTPUT", "")
     if config_dir and os.path.isdir(config_dir):
         config = os.path.join(config_dir, config)
+        makef = os.path.join(config_dir, makef)
 
     if os.path.exists(config):
         if args.no_update:
@@ -371,12 +379,23 @@ def do_it():
             return 1
 
     if maketarget is not None:
+        make_args = []
+        if not os.path.exists(makef):
+            if args.verbose:
+                sys.stderr.write(f"missing {makef}, adding -f $src/Makefile\n")
+            if os.path.exists("Makefile"):
+                # assuming we're in linux srcdir
+                make_args = ["-f", str(os.path.abspath("Makefile"))]
+                if args.verbose:
+                    sys.stderr.write(f"adding make_args: {make_args}\n")
         try:
-            subprocess.check_call(["make"] + archargs + [maketarget])
+            subprocess.check_call(["make"] + make_args + archargs + [maketarget])
         except Exception as exc:
             raise SilentError() from exc
 
     # Append virtme configs
+    if args.verbose:
+        sys.stderr.write(f"appending to config: {config}\n")
     with open(config, "ab") as conffile:
         conffile.write("\n".join(conf).encode("utf-8"))
 
