@@ -11,6 +11,7 @@ import atexit
 import argparse
 import tempfile
 import os
+import platform
 import errno
 import fcntl
 import sys
@@ -23,13 +24,14 @@ import termios
 from shutil import which
 from time import sleep
 from base64 import b64encode
+from virtme_ng.utils import CACHE_DIR
 from .. import virtmods
 from .. import modfinder
 from .. import mkinitramfs
 from .. import qemu_helpers
 from .. import architectures
 from .. import resources
-from ..util import SilentError, uname, get_username, find_binary_or_raise
+from ..util import SilentError, get_username, find_binary_or_raise
 
 
 def make_parser() -> argparse.ArgumentParser:
@@ -46,7 +48,7 @@ def make_parser() -> argparse.ArgumentParser:
         "--installed-kernel",
         action="store",
         nargs="?",
-        const=uname.release,
+        const=platform.release(),
         default=None,
         metavar="VERSION",
         help="[Deprecated] use --kimg instead.",
@@ -56,7 +58,7 @@ def make_parser() -> argparse.ArgumentParser:
         "--kimg",
         action="store",
         nargs="?",
-        const=uname.release,
+        const=platform.release(),
         default=None,
         help="Use specified kernel image or an installed kernel version. "
         + "If no argument is specified the running kernel will be used.",
@@ -74,7 +76,7 @@ def make_parser() -> argparse.ArgumentParser:
         "--mods",
         action="store",
         choices=["none", "use", "auto"],
-        default="use",
+        required=True,
         help="Setup loadable kernel modules inside a compiled kernel source directory "
         + "(used in conjunction with --kdir); "
         + "none: ignore kernel modules, use: asks user to refresh virtme's kernel modules directory, "
@@ -120,8 +122,14 @@ def make_parser() -> argparse.ArgumentParser:
         action="append",
         const="user",
         nargs="?",
-        choices=["user", "bridge", "loop"],
-        help="Enable basic network access.",
+        help="Enable basic network access: user, bridge(=<br>), loop.",
+    )
+    g.add_argument(
+        "--net-mac-address",
+        action="store",
+        default=None,
+        help="The MAC address to assign to the NIC interface, e.g. 52:54:00:12:34:56. "
+        + "The last octet will be incremented for the next network devices.",
     )
     g.add_argument(
         "--balloon",
@@ -162,6 +170,12 @@ def make_parser() -> argparse.ArgumentParser:
         default=None,
         help="Create NUMA nodes in the guest.",
     )
+    g.add_argument(
+        "--numa-distance",
+        action="append",
+        default=None,
+        help="Define a distance between two NUMA nodes in the guest (src=ID,dst=ID,val=NUM).",
+    )
     g.add_argument(
         "--cpus", action="store", default=None, help="Set guest cpu and qemu -smp flag."
     )
@@ -201,7 +215,7 @@ def make_parser() -> argparse.ArgumentParser:
         "--arch",
         action="store",
         metavar="ARCHITECTURE",
-        default=uname.machine,
+        default=platform.machine(),
         help="Guest architecture",
     )
     g.add_argument(
@@ -312,6 +326,43 @@ def make_parser() -> argparse.ArgumentParser:
         help="Supply a directory that is r/w to the guest but read-only in the host.  Use --overlay-rwdir=path.",
     )
 
+    g.add_argument(
+        "--nvgpu", action="store", default=None, help="Set guest NVIDIA GPU."
+    )
+
+    g = parser.add_argument_group(title="Remote Console")
+    cli_srv_choices = ["console", "ssh"]
+    g.add_argument(
+        "--server",
+        action="store",
+        const=cli_srv_choices[0],
+        nargs="?",
+        choices=cli_srv_choices,
+        help="Enable a server to communicate later from the host to the device using '--client'. "
+        + "By default, a simple console will be offered using a VSOCK connection, and 'socat' for the proxy."
+    )
+    g.add_argument(
+        "--client",
+        action="store",
+        const=cli_srv_choices[0],
+        nargs="?",
+        choices=cli_srv_choices,
+        help="Connect to a VM launched with the '--server' option for a remote control.",
+    )
+    g.add_argument(
+        "--port",
+        action="store",
+        type=int,
+        default=2222,
+        help="Unique port to communicate with a VM.",
+    )
+    g.add_argument(
+        "--remote-cmd",
+        action="store",
+        metavar="COMMAND",
+        help="To start in the VM a different command than the default one (--server), "
+        + "or to launch this command instead of a prompt (--client).",
+    )
     return parser
 
 
@@ -368,7 +419,7 @@ def get_rootfs_from_kernel_path(path):
     return os.path.abspath(path)
 
 
-def get_kernel_version(path):
+def get_kernel_version(img_name, path):
     if not os.path.exists(path):
         arg_fail(
             "kernel file %s does not exist, try --build to build the kernel" % path
@@ -380,7 +431,7 @@ def get_kernel_version(path):
             ["file", path], capture_output=True, text=True, check=False
         )
         for item in result.stdout.split(", "):
-            match = re.search(r"^[vV]ersion (\S+)", item)
+            match = re.search(r"^[vV]ersion (\S{3,})", item)
             if match:
                 kernel_version = match.group(1)
                 return kernel_version
@@ -393,10 +444,18 @@ def get_kernel_version(path):
     result = subprocess.run(
         ["strings", path], capture_output=True, text=True, check=False
     )
-    match = re.search(r"Linux version (\S+)", result.stdout)
+    match = re.search(r"Linux version (\S{3,})", result.stdout)
     if match:
         kernel_version = match.group(1)
         return kernel_version
+
+    # The version detection fails s390x using file or strings tools, so check
+    # if the file itself contains the version number.
+    if img_name:
+        match = re.search(fr"{img_name}-(\S{{3,}})", path)
+        if match:
+            return match.group(1)
+
     return None
 
 
@@ -428,16 +487,18 @@ def find_kernel_and_mods(arch, args) -> Kernel:
                 args.kimg = None
 
     if args.kimg is not None:
+        img_name = arch.img_name()
+
         # Try to resolve kimg as a kernel version first, then check if a file
         # is provided.
-        kimg = "/usr/lib/modules/%s/vmlinuz" % args.kimg
+        kimg = "/usr/lib/modules/%s/%s" % (args.kimg, img_name)
         if not os.path.exists(kimg):
-            kimg = "/boot/vmlinuz-%s" % args.kimg
+            kimg = "/boot/%s-%s" % (img_name, args.kimg)
             if not os.path.exists(kimg):
                 kimg = args.kimg
                 if not os.path.exists(kimg):
                     arg_fail("%s does not exist" % args.kimg)
-        kver = get_kernel_version(kimg)
+        kver = get_kernel_version(img_name, kimg)
         if kver is None:
             # Unable to detect kernel version, try to boot without
             # automatically detecting modules.
@@ -484,7 +545,7 @@ def find_kernel_and_mods(arch, args) -> Kernel:
     elif args.kdir is not None:
         kimg = os.path.join(args.kdir, arch.kimg_path())
         # Run get_kernel_version to check at least if the kernel image exist.
-        kernel.version = get_kernel_version(kimg)
+        kernel.version = get_kernel_version(None, kimg)
         kernel.kimg = kimg
         virtme_mods = os.path.join(args.kdir, ".virtme_mods")
         mod_file = os.path.join(args.kdir, "modules.order")
@@ -573,7 +634,7 @@ class VirtioFS:
     def _get_virtiofsd_path(self):
         # Define the possible virtiofsd paths.
         #
-        # NOTE: do not use the C implemention of qemu's virtiofsd, because it
+        # NOTE: do not use the C implementation of qemu's virtiofsd, because it
         # doesn't support unprivileged-mode execution and it would be totally
         # unsafe to export the whole rootfs of the host running as root.
         #
@@ -652,48 +713,59 @@ class VirtioFS:
         return True
 
 
+class VirtioFSConfig:
+    def __init__(self, path: str, mount_tag: str, guest_tools_path=None, memory=None):
+        self.path = path
+        self.mount_tag = mount_tag
+        self.guest_tools_path = guest_tools_path
+        self.memory = memory
+
+
 def export_virtiofs(
     arch: architectures.Arch,
     qemuargs: List[str],
-    path: str,
-    mount_tag: str,
-    guest_tools_path=None,
-    memory=None,
+    config: VirtioFSConfig,
     verbose=False,
-) -> None:
+) -> bool:
     if not arch.virtiofs_support():
         return False
 
-    # Try to start virtiofsd deamon
-    virtio_fs = VirtioFS(guest_tools_path)
-    ret = virtio_fs.start(path, verbose)
+    # Try to start virtiofsd daemon
+    virtio_fs = VirtioFS(config.guest_tools_path)
+    ret = virtio_fs.start(config.path, verbose)
     if not ret:
         return False
 
     # Adjust qemu options to use virtiofsd
     fsid = "virtfs%d" % len(qemuargs)
-
     vhost_dev_type = arch.vhost_dev_type()
+
     qemuargs.extend(["-chardev", f"socket,id=char{fsid},path={virtio_fs.sock}"])
-    qemuargs.extend(["-device", f"{vhost_dev_type},chardev=char{fsid},tag={mount_tag}"])
-    if memory is None:
-        memory = "128M"
-    elif memory == 0:
+    qemuargs.extend(["-device", f"{vhost_dev_type},chardev=char{fsid},tag={config.mount_tag}"])
+
+    memory = config.memory if config.memory is not None else "128M"
+    if memory == 0:
         return True
+
     qemuargs.extend(["-object", f"memory-backend-memfd,id=mem,size={memory},share=on"])
     qemuargs.extend(["-numa", "node,memdev=mem"])
 
     return True
 
 
+class VirtFSConfig:
+    def __init__(self, path: str, mount_tag: str, security_model="none", readonly=True):
+        self.path = path
+        self.mount_tag = mount_tag
+        self.security_model = security_model
+        self.readonly = readonly
+
+
 def export_virtfs(
     qemu: qemu_helpers.Qemu,
     arch: architectures.Arch,
     qemuargs: List[str],
-    path: str,
-    mount_tag: str,
-    security_model="none",
-    readonly=True,
+    config: VirtFSConfig,
 ) -> None:
     # NB: We can't use -virtfs for this, because it can't handle a mount_tag
     # that isn't a valid QEMU identifier.
@@ -704,9 +776,9 @@ def export_virtfs(
             "local,id=%s,path=%s,security_model=%s%s%s"
             % (
                 fsid,
-                qemu.quote_optarg(path),
-                security_model,
-                ",readonly=on" if readonly else "",
+                qemu.quote_optarg(config.path),
+                config.security_model,
+                ",readonly=on" if config.readonly else "",
                 ",multidevs=remap" if qemu.has_multidevs else "",
             ),
         ]
@@ -715,7 +787,7 @@ def export_virtfs(
         [
             "-device",
             "%s,fsdev=%s,mount_tag=%s"
-            % (arch.virtio_dev_type("9p"), fsid, qemu.quote_optarg(mount_tag)),
+            % (arch.virtio_dev_type("9p"), fsid, qemu.quote_optarg(config.mount_tag)),
         ]
     )
 
@@ -808,6 +880,131 @@ def is_subpath(path, potential_parent):
     return common_path == potential_parent
 
 
+def get_console_path(port):
+    return os.path.join(tempfile.gettempdir(), "virtme-console", f"{port}.sh")
+
+
+def console_client(args):
+    try:
+        # with tty support
+        (cols, rows) = os.get_terminal_size()
+        stty = f'stty rows {rows} cols {cols} iutf8 echo'
+        socat_in = f'file:{os.ttyname(sys.stdin.fileno())},raw,echo=0'
+    except OSError:
+        stty = ''
+        socat_in = '-'
+    socat_out = f'VSOCK-CONNECT:{args.port}:1024'
+
+    user = args.user if args.user else '${virtme_user:-root}'
+
+    if args.pwd:
+        cwd = os.path.relpath(os.getcwd(), args.root)
+    elif args.cwd is not None:
+        cwd = os.path.relpath(args.cwd, args.root)
+    else:
+        cwd = '${virtme_chdir:+"${virtme_chdir}"}'
+
+    # use 'su' only if needed: another use, or to get a prompt
+    cmd = f'if [ "{user}" != "root" ]; then\n' + \
+          f'  exec su "{user}"'
+    if args.remote_cmd is not None:
+        exec_escaped = args.remote_cmd.replace('"', '\\"')
+        cmd += f' -c "{exec_escaped}"' + \
+               '\nelse\n' + \
+               f'  {args.remote_cmd}\n'
+    else:
+        cmd += '\nelse\n' + \
+                '  exec su\n'
+    cmd += 'fi'
+
+    console_script_path = get_console_path(args.port)
+    with open(console_script_path, 'w', encoding="utf-8") as file:
+        print((
+            '#! /bin/bash\n'
+            'main() {\n'
+            f'{stty}\n'
+            f'HOME=$(getent passwd "{user}" | cut -d: -f6)\n'
+            f'cd {cwd}\n'
+            f'{cmd}\n'
+            '}\n'
+            'main'  # use a function to avoid issues when the script is modified
+        ), file=file)
+    os.chmod(console_script_path, 0o755)
+
+    if args.dry_run:
+        print('socat', socat_in, socat_out)
+    else:
+        os.execvp('socat', ['socat', socat_in, socat_out])
+
+
+def console_server(args, qemu, arch, qemuargs, kernelargs):
+    console_script_path = get_console_path(args.port)
+    if os.path.exists(console_script_path):
+        arg_fail("console: '%s' file exists: " % console_script_path
+                 + "another VM is running with the same --port? "
+                 + "If not, remove this file.")
+
+    def cleanup_console_script():
+        os.unlink(console_script_path)
+
+    # create an empty file that can be populated later on
+    console_script_dir = os.path.dirname(console_script_path)
+    os.makedirs(console_script_dir, exist_ok=True)
+    open(console_script_path, 'w', encoding="utf-8").close()
+    atexit.register(cleanup_console_script)
+
+    if args.remote_cmd is not None:
+        console_exec = args.remote_cmd
+    else:
+        console_exec = console_script_path
+        if args.root != "/":
+            virtfs_config = VirtFSConfig(
+                path=console_script_dir,
+                mount_tag="virtme.vsockmount",
+            )
+            export_virtfs(qemu, arch, qemuargs, virtfs_config)
+            kernelargs.append("virtme_vsockmount=%s" % console_script_dir)
+
+    kernelargs.extend([f"virtme.vsockexec=`{console_exec}`"])
+    qemuargs.extend(["-device", "vhost-vsock-pci,guest-cid=%d" % args.port])
+
+
+def ssh_client(args):
+    if args.remote_cmd is not None:
+        exec_escaped = args.remote_cmd.replace('"', '\\"')
+        remote_cmd = ["bash", "-c", exec_escaped]
+    else:
+        remote_cmd = []
+
+    cmd = ["ssh", "-p", str(args.port), "localhost"] + remote_cmd
+
+    if args.dry_run:
+        print(" ".join(cmd))
+    else:
+        os.execvp("ssh", cmd)
+
+
+def ssh_server(args, arch, qemuargs, kernelargs):
+    # Check if we need to generate the ssh host keys for the guest.
+    ssh_key_dir = f"{CACHE_DIR}/.ssh"
+    os.makedirs(f"{ssh_key_dir}/etc/ssh", exist_ok=True)
+    os.system(f"ssh-keygen -A -f {ssh_key_dir}")
+
+    # Implicitly enable dhcp to automatically get an IP on the network
+    # interface and prevent interface renaming.
+    kernelargs.extend(["virtme.dhcp", "net.ifnames=0", "biosdevname=0"])
+
+    # Tell virtme-ng-init / virtme-init to start sshd and use the current
+    # username keys/credentials.
+    username = get_username()
+    kernelargs.extend(["virtme.ssh"])
+    kernelargs.extend([f"virtme_ssh_user={username}"])
+
+    # Setup a port forward network interface for the guest.
+    qemuargs.extend(["-device", "%s,netdev=ssh" % (arch.virtio_dev_type("net"))])
+    qemuargs.extend(["-netdev", "user,id=ssh,hostfwd=tcp::%d-:22" % args.port])
+
+
 # Allowed characters in mount paths.  We can extend this over time if needed.
 _SAFE_PATH_PATTERN = "[a-zA-Z0-9_+ /.-]+"
 _RWDIR_RE = re.compile("^(%s)(?:=(%s))?$" % (_SAFE_PATH_PATTERN, _SAFE_PATH_PATTERN))
@@ -816,8 +1013,18 @@ _RWDIR_RE = re.compile("^(%s)(?:=(%s))?$" % (_SAFE_PATH_PATTERN, _SAFE_PATH_PATT
 def do_it() -> int:
     args = _ARGPARSER.parse_args()
 
+    if args.client is not None:
+        if args.server is not None:
+            arg_fail('--client cannot be used with --server.')
+
+        if args.client == 'console':
+            console_client(args)
+        elif args.client == 'ssh':
+            ssh_client(args)
+        sys.exit(0)
+
     arch = architectures.get(args.arch)
-    is_native = args.arch == uname.machine
+    is_native = args.arch == platform.machine()
 
     qemu = qemu_helpers.Qemu(args.qemu_bin, arch.qemuname)
     qemu.probe()
@@ -864,9 +1071,12 @@ def do_it() -> int:
     # Propagate /proc/sys/fs/nr_open from the host to the guest, otherwise we
     # may see some EPERM errors, because certain applications/settings may
     # expect to be able to use a higher limit of the max number of open files.
-    with open('/proc/sys/fs/nr_open', 'r', encoding="utf-8") as file:
-        nr_open = file.readline().strip()
-        kernelargs.append(f"nr_open={nr_open}")
+    try:
+        with open('/proc/sys/fs/nr_open', 'r', encoding="utf-8") as file:
+            nr_open = file.readline().strip()
+            kernelargs.append(f"nr_open={nr_open}")
+    except FileNotFoundError:
+        pass
 
     # Parse NUMA settings.
     if args.numa:
@@ -878,6 +1088,12 @@ def do_it() -> int:
                 "-numa", f"node,memdev=mem{i}{cpus}"
             ])
 
+    if args.numa_distance:
+        for arg in args.numa_distance:
+            qemuargs.extend([
+                "-numa", f"dist,{arg}"
+            ])
+
     if args.snaps:
         if args.root == "/":
             snapd_state = "/var/lib/snapd/state.json"
@@ -926,16 +1142,19 @@ def do_it() -> int:
             virt_arch = architectures.get("microvm")
         else:
             virt_arch = arch
-        use_virtiofs = export_virtiofs(
-            virt_arch,
-            qemuargs,
-            args.root,
-            "ROOTFS",
+        virtiofs_config = VirtioFSConfig(
+            path=args.root,
+            mount_tag="ROOTFS",
             guest_tools_path=guest_tools_path,
             # virtiofsd requires a NUMA not, if --numa is specified simply use
             # the user-defined NUMA node, otherwise create a NUMA node with all
             # the memory.
             memory=0 if args.numa else args.memory,
+        )
+        use_virtiofs = export_virtiofs(
+            virt_arch,
+            qemuargs,
+            virtiofs_config,
             verbose=args.verbose,
         )
         if can_use_microvm(args) and use_virtiofs:
@@ -943,9 +1162,12 @@ def do_it() -> int:
                 sys.stderr.write("virtme: use 'microvm' QEMU architecture\n")
             arch = virt_arch
     if not use_virtiofs:
-        export_virtfs(
-            qemu, arch, qemuargs, args.root, "/dev/root", readonly=(not args.rw)
+        virtfs_config = VirtFSConfig(
+            path=args.root,
+            mount_tag="/dev/root",
+            readonly=(not args.rw),
         )
+        export_virtfs(qemu, arch, qemuargs, virtfs_config)
 
     # Use the faster virtme-ng-init if we are running on a native architecture.
     if (
@@ -960,7 +1182,11 @@ def do_it() -> int:
     if args.root == "/":
         initcmds = [f"init={guest_tools_path}/{virtme_init_cmd}"]
     else:
-        export_virtfs(qemu, arch, qemuargs, guest_tools_path, "virtme.guesttools")
+        virtfs_config = VirtFSConfig(
+            path=guest_tools_path,
+            mount_tag="virtme.guesttools",
+        )
+        export_virtfs(qemu, arch, qemuargs, virtfs_config)
         initcmds = [
             "init=/bin/sh",
             "--",
@@ -1027,9 +1253,12 @@ def do_it() -> int:
         idx = mount_index
         mount_index += 1
         tag = "virtme.initmount%d" % idx
-        export_virtfs(
-            qemu, arch, qemuargs, hostpath, tag, readonly=(dirtype != "rwdir")
+        virtfs_config = VirtFSConfig(
+            path=hostpath,
+            mount_tag=tag,
+            readonly=(dirtype != "rwdir"),
         )
+        export_virtfs(qemu, arch, qemuargs, virtfs_config)
         kernelargs.append("virtme_initmount%d=%s" % (idx, guestpath))
 
     for i, d in enumerate(args.overlay_rwdir):
@@ -1037,11 +1266,14 @@ def do_it() -> int:
 
     # Turn on KVM if available
     kvm_ok = can_use_kvm(args)
-    if is_native and kvm_ok:
-        qemuargs.extend(["-machine", "accel=kvm:tcg"])
+    if is_native:
+        if kvm_ok:
+            qemuargs.extend(["-machine", "accel=kvm:tcg"])
+        elif platform.system() == "Darwin":
+            qemuargs.extend(["-machine", "accel=hvf"])
 
     # Add architecture-specific options
-    qemuargs.extend(arch.qemuargs(is_native, kvm_ok))
+    qemuargs.extend(arch.qemuargs(is_native, kvm_ok, args.nvgpu is not None))
 
     # Set up / override baseline devices
     qemuargs.extend(["-parallel", "none"])
@@ -1077,7 +1309,10 @@ def do_it() -> int:
 
         kernelargs.extend(["virtme_console=" + arg for arg in arch.serial_console_args()])
 
-        qemuargs.extend(arch.qemu_nodisplay_args())
+        if args.nvgpu is None:
+            qemuargs.extend(arch.qemu_nodisplay_args())
+        else:
+            qemuargs.extend(arch.qemu_nodisplay_nvgpu_args())
 
         # PS/2 probing is slow; give the kernel a hint to speed it up.
         kernelargs.extend(["psmouse.proto=exps"])
@@ -1162,7 +1397,10 @@ def do_it() -> int:
     def do_script(shellcmd: str, ret_path=None, show_boot_console=False) -> None:
         if args.graphics is None:
             # Turn off default I/O
-            qemuargs.extend(arch.qemu_nodisplay_args())
+            if args.nvgpu is None:
+                qemuargs.extend(arch.qemu_nodisplay_args())
+            else:
+                qemuargs.extend(arch.qemu_nodisplay_nvgpu_args())
 
         # Check if we can redirect stdin/stdout/stderr.
         if not can_access_file("/proc/self/fd/0") or \
@@ -1275,30 +1513,47 @@ def do_it() -> int:
             show_boot_console=args.show_boot_console,
         )
 
-    if args.graphics is not None:
+    if args.graphics is not None and args.nvgpu is None:
         video_args = arch.qemu_display_args()
         if video_args:
             qemuargs.extend(video_args)
 
+    def get_net_mac(index):
+        if args.net_mac_address is None:
+            return ""
+
+        mac = args.net_mac_address.split(':')
+        try:
+            mac[5] = "%02x" % ((int(mac[5], 16) + index) % 256)
+        except (ValueError, IndexError):
+            arg_fail("--net-mac-address: invalid MAC address: '%s'" % args.net_mac_address)
+        return ",mac=" + ":".join(mac)
+
     if args.net:
         extend_dhcp = False
         index = 0
         for net in args.net:
-            qemuargs.extend(["-device", "%s,netdev=n%d" % (arch.virtio_dev_type("net"), index)])
+            qemuargs.extend(["-device", "%s,netdev=n%d%s" %
+                            (arch.virtio_dev_type("net"), index, get_net_mac(index))])
             if net == "user":
                 qemuargs.extend(["-netdev", "user,id=n%d" % index])
                 extend_dhcp = True
-            elif net == "bridge":
-                qemuargs.extend(["-netdev", "bridge,id=n%d,br=virbr0" % index])
+            elif net == "bridge" or net.startswith("bridge="):
+                if len(net) > 7 and net[6] == '=':
+                    bridge = net[7:]
+                else:
+                    bridge = "virbr0"
+                qemuargs.extend(["-netdev", "bridge,id=n%d,br=%s" % (index, bridge)])
                 extend_dhcp = True
             elif net == "loop":
                 hubid = index
                 qemuargs.extend(["-netdev", "hubport,id=n%d,hubid=%d" % (index, hubid)])
                 index += 1
-                qemuargs.extend(["-device", "%s,netdev=n%d" % (arch.virtio_dev_type("net"), index)])
+                qemuargs.extend(["-device", "%s,netdev=n%d%s" %
+                                (arch.virtio_dev_type("net"), index, get_net_mac(index))])
                 qemuargs.extend(["-netdev", "hubport,id=n%d,hubid=%d" % (index, hubid)])
             else:
-                assert False
+                arg_fail("--net: invalid choice: '%s' (choose from user, bridge(=<br>), loop)" % net)
             index += 1
         if extend_dhcp:
             kernelargs.extend(["virtme.dhcp"])
@@ -1310,6 +1565,12 @@ def do_it() -> int:
             ]
         )
 
+    if args.server is not None:
+        if args.server == "console":
+            console_server(args, qemu, arch, qemuargs, kernelargs)
+        elif args.server == "ssh":
+            ssh_server(args, arch, qemuargs, kernelargs)
+
     if args.pwd:
         rel_pwd = os.path.relpath(os.getcwd(), args.root)
         if rel_pwd.startswith(".."):
@@ -1326,9 +1587,12 @@ def do_it() -> int:
             return 1
         kernelargs.append("virtme_chdir=%s" % rel_cwd)
 
-    if args.user:
+    if args.user and args.user != "root":
         kernelargs.append("virtme_user=%s" % args.user)
 
+    if args.nvgpu:
+        qemuargs.extend(["-device", args.nvgpu])
+
     # If we are running as root on the host pass this information to the guest
     # (this can be useful to properly support running virtme-ng instances
     # inside docker)
